#ifndef TOOLKITS_SYSTEMTK_H_
#define TOOLKITS_SYSTEMTK_H_

/**
 * Toolkit for system functions.
 */
class SystemTk
{
	public:
		static std::string getUsername();

	private:
		SystemTk() {}
};



#endif /* TOOLKITS_SYSTEMTK_H_ */
